#include "hacks.h"
#ifndef MAX_STR
  #define MAX_STR 255
#endif


/* global vars */
int master_file_size = 79837557; /* this will change with each patch */
char master_file_name[] = "morrowind.esm";


/* forward references */
void usage();
void usage_exit();
int read_intel_long(FILE *, int *);
int write_intel_long(FILE *, int);
int write_intel_float(FILE *, float);
int read_line(FILE *, char *);
void trim_spaces(char *);
int write_nulls(FILE *, int);
int write_esp_header(FILE *);
int read_esp_header(FILE *);
int copy_file(FILE *, FILE *);
int copy_cell_from_file(char *, FILE *);


/* start of actual program */
void usage()
{
  fprintf(stderr, "espjoin\n\
program to join a bunch of ESP files.  Important note: this program only\n\
works with the subset of the ESP format that is created by the txt2esp4\n\
program\n");
  fprintf(stderr, "\
Usage: espjoin <filelist> <outfile>\nwhere file list is a text file\n\
containing the list of esp files to join, one file per line\n");
}


void usage_exit()
{
  usage();
  exit(DOOPS);
}


int read_intel_long(f, n)
FILE *f;
int *n;
{
  int status;

  status = fread( (char *) n, 4, 1, f);
  if (status neq 1)
    return(DOOPS);
  return(OKDOKEY);
}


int write_intel_long(f, n)
FILE *f;
int n;
{
  int status;

  status = fwrite( (char *) &n, 4, 1, f);
  if (status neq 1)
    return(DOOPS);
  return(OKDOKEY);
}


int write_intel_float(f, fL)
FILE *f;
float fL;
{
  int status;

  /* we got lucky.  in this compiler, it uses the same float format
     as the esm file.  so no translation is needed at all. */
  status = fwrite( (char *) &fL, 4, 1, f);
  if (status neq 1)
    return(DOOPS);
  return(OKDOKEY);
}


/* read_line
   will read a complete line, all the way to the end of line.
   if the line is longer than the string, the string will be truncated
   but the file pointer will be pointing to the first character of the
   next line */
int read_line(f, s)
FILE *f;
char *s;
{
  int i,c;

  for (i = 0;;i++) {
    if ((c = fgetc(f)) eq EOF) break;
    if (c eq '\n') break;
    if ((i - 1) < MAX_STR) {
      s[i] = c;
    }
  }
  if (i >= MAX_STR)
    i = (MAX_STR - 1);
  s[i] = 0;
  if ((c eq EOF) && (i eq 0))
    return(c);
  return(OKDOKEY);
}


void trim_spaces(s)
char *s;
{
  int lead_s, end_s, s_len;

  s_len = strlen(s);
  if (s_len) {
    /* first deal with leading space */
    lead_s = 0;
    while (lead_s < s_len) {
      if (s[lead_s] > ' ')
        break;
      lead_s++;
    }
    memmove(s, &s[lead_s], 1 + (s_len - lead_s)); /* will copy null if zero len str */
    /* leading space gone.  now see to the trailing spaces */
    end_s = strlen(s);
    if (end_s > 0) {
      while (end_s-- > 0) {
        if (s[end_s] > ' ') { /* if non space .... */
          s[++end_s] = '\0';  /* put null on end of string */
          end_s = 0;          /* signal for end of loop */
        }
      }
    }
  } /* end if s_len */
}


int write_nulls(f, how_many)
FILE *f;
int how_many;
{
  int status;

  while(how_many--)
    if ((status = fputc(0, f)) eq EOF)
      return(DOOPS);
}


int write_esp_header(f)
FILE *f;
{
  int status, temp4, temp5;
  float ftemp;

  fprintf(f, "TES3");
  status = write_intel_long(f, 332 + strlen(master_file_name) + 1);
  if (status eq DOOPS) return(status);
  /* write 8 nulls */
  status = write_nulls(f, 8);
  if (status eq DOOPS) return(status);
  fprintf(f, "HEDR");
  status = write_intel_long(f, 300);
  if (status eq DOOPS) return(status);
  /* version number of esm file is a four byte float.
     This is set to the 1.2 (patched) version */
  status = write_intel_float(f, 1.2);
  if (status eq DOOPS) return(status);
  status = write_nulls(f, 4);
  if (status eq DOOPS) return(status);
  /* author name, 32 chars, all nulls for now */
  status = write_nulls(f, 32);  
  if (status eq DOOPS) return(status);
  /* description of file, 260 chars, all nulls for now */
  status = write_nulls(f, 260);  
  if (status eq DOOPS) return(status);
  fprintf(f, "MAST");
  /* length of master filename including null */
  status = write_intel_long(f, strlen(master_file_name) + 1);
  if (status eq DOOPS) return(status);
  fprintf(f, master_file_name);
  status = fputc(0, f); /* write null terminator */
  if (status eq EOF) return(status);
  fprintf(f, "DATA");
  status = write_intel_long(f, 8);
  if (status eq DOOPS) return(status);
  status = write_intel_long(f, master_file_size); /* low four bytes */
  if (status eq DOOPS) return(status);
  status = write_intel_long(f, 0);                /* high four bytes */
  if (status eq DOOPS) return(status);
  return(OKDOKEY);
}


/* read past esp header and have file pointer at first non-header byte */
int read_esp_header(f)
FILE *f;
{
  int status;
  int header_size; /* int had better be 4 bytes in your ccompiler */
  char temp[4];
  
  /* read first four butes - they should be "TES3" */
  if ((status = fread(&temp[0], 4, 1, f)) neq 1) {
    fprintf(stderr, "read error in read_esp_header\n");
    return(DOOPS);
  }
  status = 0;
  if (temp[0] neq 'T') status++;
  if (temp[1] neq 'E') status++;
  if (temp[2] neq 'S') status++;
  if (temp[3] neq '3') status++;
  if (status) {
    fprintf(stderr, "not a TES3 esp file\n");
    return(DOOPS);
  }
  /* read total header size */
  if (
       (status = read_intel_long(f, &header_size))
       eq
       DOOPS
     ) {
    fprintf(stderr, "read_esp_header, error getting header size\n");
    return(status);
  }
  /* skip past rest of header */
  status = fseek(f, (8 + header_size), SEEK_CUR);
  if (status neq 0) {
    fprintf(stderr, "read_esp_header: seek error\n");
    return(DOOPS);
  }
  return(OKDOKEY);
}


int copy_file(f1, f2)
FILE *f1;
FILE *f2;
{
  int c, status;

  while ((c = fgetc(f1)) neq EOF) {
    if ((status = fputc(c, f2)) eq EOF) {
      fprintf(stderr, "copy_files: file write error\n");
      return(DOOPS);
    }
  }
  return(OKDOKEY);
}


int copy_cell_from_file(in_file_name, f2)
char *in_file_name;
FILE *f2;
{
  FILE *f1;
  int status;

  f1 = fopen(in_file_name, "rb");
  if (f1 eq NULL) {
    fprintf(stderr, "error opening input esp file \"%s\"\n", in_file_name);
    return(DOOPS);
  }
  /* skip past esp header to first cell header */
  if ((status = read_esp_header(f1)) eq DOOPS) {
    return(status);
  }
  /* blindly (with no checking) copy the rest of the file */
  if ((status = copy_file(f1, f2)) eq DOOPS) {
    return(status);
  }
  /* close input file */
  fclose(f1);
  return(OKDOKEY);
}


main(argc, argv)
int argc;
char *argv[];
{
  FILE *listfile, *outfile;
  char in_esp_name[MAX_STR];
  int status;

  if (argc eq 0)
    exit(DOOPS);
  if (argc neq 3)
    usage_exit();
  if ((listfile = fopen(argv[1], "r")) eq NULL) {
    fprintf(stderr, "Error opening list file \"%s\"\n", argv[1]);
    exit(DOOPS);
  }
  if ((outfile = fopen(argv[2], "wb")) eq NULL) {
    fprintf(stderr, "Error creating output esp file \"%s\"\n", argv[2]);
    exit(DOOPS);
  }
  if ((status = write_esp_header(outfile)) neq OKDOKEY) {
    fprintf(stderr, "error writing ESP header to output file\n");
    exit(DOOPS);
  }
  /* main loop */
  while ((status = read_line(listfile, &in_esp_name[0])) neq EOF) {
    status = copy_cell_from_file(in_esp_name, outfile);
    if (status eq DOOPS) {
      fprintf(stderr, "error reading esp file \"%s\"\n", in_esp_name);
      exit(DOOPS);
    }
  }
  /* all copied.  What is really handy is that the ESP header does not
     have the size of the entire file in it, so we don't have to rewind
     and update the header.  We are done but for the closing of the
     files and the exiting. */
  fclose(outfile);
  fclose(listfile);
  exit(OKDOKEY);
}


/* actual end of this file */
