/* imperial-style dungeon generator */
/* v7 adds a different random number genertaor and uses much shorter temp
   file names.  This allows a longer base cell name */
/* v6 adds the -l option for vertical levels.  trap doors are placed in
   ceiling and floor to connect levels.  The -sl option is used to set
   the starting level.  Also the "delete_temps" option has been added. */

#include "hacks.h"

#ifndef MAX_STR
  #define MAX_STR 256
#endif
#include <errno.h>
#define maze_size 11


char include_esp_file_name[MAX_STR];
char out_bat_file_name[MAX_STR];
char out_list_file_name[MAX_STR];

char *leveled_monster[40];
char *crate[40];
char *bag[40];
char *small_chest[40];
char txtr_output_file[MAX_STR];
int leveled_monster_index = 0;
int crate_index = 0;
int bag_index = 0;
int small_chest_index = 0;
int start_point;
int save_m_x, save_m_y;
char base_file_name[MAX_STR];
char base_cell_name[MAX_STR];
char bat_output_file[MAX_STR];
char seed_file_name[] = "seedhold.dat";

int maze[maze_size][maze_size];
int rooms[maze_size][maze_size];
int room_doors_x[6];
int room_doors_y[6];
int room_doors_index = 0;
int room_corner_coords_x[6];  /* three rooms, two corners */
int room_corner_coords_y[6];
FILE *_txtr_file = (FILE *) NULL;


int x_pos, y_pos;
int xx, yy, zz;
int x_size, y_size;
int room_area;
int try_x, try_y;
int filled_neighbor;
int skip_this_door;
int how_twisty;
int leave_it;
int no_monsters, no_treasure, no_clutter, no_twisty, num_levels, lev_offset;
int del_files = 0;
int cur_level, last_dx, last_dy, new_dx, new_dy;


/* for random number generator */
unsigned char S[256];
unsigned char K[256];
unsigned int I, J;


/* forward references */
void repeat_str(int);
void initialize_thrash(int);
int ROAi(int);
void RTRT(int *, int, int, int, int);
int read_line(FILE *, char *);
void trim_spaces(char *);
void fill_defaults();
int parse_files(FILE *);
int parse_monsters(FILE *);
int parse_crate(FILE *);
int parse_bag(FILE *);
int parse_small_chest(FILE *);
int parse_ini();
void test_parse();
void usage();
void usage_exit();
void m_str(int, char *);
void cell_name(int, int, char *);
int esp_grid(int);
void place_rug(int, int, int);
void place_doorjam(int, int, int, int);
void place_door(int, int, int, int);
void place_load_door(int, int, int, int, int, int, int);
void place_tapestry(int, int, int, int);
void place_monster(int, int, int);
void place_4way(int, int, int);
void place_straight(int, int, int, int);
void place_corner(int, int, int, int);
void place_3way(int, int, int, int);
void place_cap(int, int, int, int);
void place_corridor(int, int, int);
void place_room_entry(int, int, int, int);
void place_room_corner(int, int, int, int);
void place_room_concave_corner(int, int, int, int);
void place_room_center(int, int, int);
void place_room_side(int, int, int, int);
void place_room_wonky1(int, int, int);
void place_room_wonky2(int, int, int);
void place_room(int, int, int);
int check_door(int, int);
void find_doors(int *, int *);
int write_to_line(char *, FILE *);
void build_cell(int, int, int);
int write_bat_file(char *, int, char *);


/* start of actual program */

void repeat_str(len)
int len;   /* length of seed string, already in K array */
{
  int temp, temp2;

  /* repeat the string over & over until the array is filled */
  temp = 0;
  temp2 = len;
  for (;len < 256; len++)
    K[temp2++] = K[temp++];
}


void initialize_thrash(len)
int len;  /* length of seed in bytes */
{
  unsigned int i;
  unsigned char swap;

  repeat_str(len);  /* seed has already been put in the K array */
  I = J = 0;
  for (i = 0; i neq 256; i++)
    S[i] = i;
  for (i = 0; i neq 256; i++) {
    J = (J + S[i] + K[i]) mod 256;
    swap = S[i];
    S[i] = S[J];
    S[J] = swap;
  }
  /* it is unclear whether or not it would be better
     or worse to re-initialize J to be zero again
  */
  J = 0;
  /* might as well clear the key since we're done with it */
  for (i = 0; i neq 256; i++)
    K[i] = 0;
}


int ROAi(the_i)  /* Read Only Acess to integer, a forcive kludge */
int the_i;
{
  int temp;

  temp = the_i;
  return(temp);
}


void RTRT(local_trans, rot, xx, yy, zz)
int *local_trans;  /* array of 4, in rot, x, y, z, order */
int rot;  /* must be 0, 90, 180, 270 ONLY */
int xx;
int yy;
int zz;
{
  int temp;
  switch (rot) {
    case 90 :
      temp = local_trans[1];
      local_trans[1] = local_trans[2];
      local_trans[2] = 0 - temp;
      break;
    case 180 :
      local_trans[1] = 0 - local_trans[1];
      local_trans[2] = 0 - local_trans[2];
      break;
    case 270 :
      temp = local_trans[1];
      local_trans[1] = 0 - local_trans[2];
      local_trans[2] = temp;
      break;
  }  /* end switch */
  local_trans[1] += xx;
  local_trans[2] += yy;
  local_trans[3] += zz;
  local_trans[0] = (local_trans[0] + rot) mod 360;
}


unsigned char get_rnd_byte()
{
  unsigned int temp;
  unsigned char swap;

  I++;
  I = I mod 256;   /* same as &= 255 but hopefully more portable */
  J = (J + S[I]) mod 256;
  swap = S[I];
  S[I] = S[J];
  S[J] = swap;
  temp = S[I] + S[J];
  temp = temp mod 256;
  return( S[temp] );
}


void set_seed()
{
  unsigned int time_value;  /* had better be 4 bytes */
  unsigned int x;
  unsigned char dummy;
  int status;
  FILE *seed_file;

  if ((seed_file = fopen(seed_file_name, "rb")) eq NULL) {
    /* seed file does not exist.  Just quietly do without it */
    x = 0;
  } else {  /* have seed file, so use it */
    status = fread(&x, sizeof(int), 1, seed_file);
    if (status eq 0) { /* read was not OK - do without it */
      fprintf(stderr, "warning: error reading seed hold file\n");
      x = 0;
    }
    fclose(seed_file);
  }
  time_value = time(NULL);    /* get time */
  memmove(&K[0], &time_value, 4);  /* assumes ints are always 4 bytes */
  memmove(&K[4], &x, 4);  /* assumes ints are always 4 bytes */
  initialize_thrash(8);  /* initialize random number generator */

  x = 1025;
  while (x--)  /* warm up rnd number generator */
    dummy = get_rnd_byte();
}


void update_seed()
{
  unsigned long new_seed;
  int status;
  FILE *seed_file;

  status = get_rnd_byte();
  new_seed = (unsigned long) get_rnd_byte();
  new_seed *= 256;
  status = get_rnd_byte();
  new_seed += (unsigned long) get_rnd_byte();
  new_seed *= 256;
  status = get_rnd_byte();
  new_seed += (unsigned long) get_rnd_byte();
  new_seed *= 256;
  status = get_rnd_byte();
  new_seed += (unsigned long) get_rnd_byte();

  seed_file = fopen(seed_file_name, "wb");
  if (status eq NULL) {
    fprintf(stderr, "warning: error creating seed hold file\n");
  } else {
    status = fwrite(&new_seed, sizeof(unsigned long), 1, seed_file);
    if (status neq 1) { /* write error */
      fprintf(stderr, "warning: error writing seed hold file\n");
      fclose(seed_file);
      remove(seed_file_name); /* delete the file if you can */
    } else { /* no write error */
      fclose(seed_file);
    }
  }
}


long rnd(r)
long r;
{
  unsigned long temp;
  unsigned char *ptr;

  ptr = (char *) &temp;
  /* a long is four bytes */
  *(ptr++) = get_rnd_byte();
  *(ptr++) = get_rnd_byte();
  *(ptr++) = get_rnd_byte();
  *ptr = get_rnd_byte();
  ptr = (char *) NULL;  /* discard pointer */
  temp = temp mod r;  /* not statistically pure, but adequate */
  return((long) temp);
}


int write_intel_long(f, n)
FILE *f;
int n;
{
  int status;

  status = fwrite( (char *) &n, 4, 1, f);
  if (status neq 1)
    return(DOOPS);
  return(OKDOKEY);
}



/* read_line
   will read a complete line, all the way to the end of line.
   if the line is longer than the string, the string will be truncated
   but the file pointer will be pointing to the first character of the
   next line */
int read_line(f, s)
FILE *f;
char *s;
{
  int i,c;

  for (i = 0;; i++) {
    if ((c = fgetc(f)) eq EOF) break;
    if (c eq '\n') break;
    if ((i - 1) < MAX_STR) {
      s[i] = c;
    }
  }
  if (i >= MAX_STR)
    i = (MAX_STR - 1);
  s[i] = 0;
  if ((c eq EOF) && (i eq 0))
    return(c);
  return(OKDOKEY);
}


void trim_spaces(s)
char *s;
{
  int lead_s, end_s, s_len;

  s_len = strlen(s);
  if (s_len) {
    /* first deal with leading space */
    lead_s = 0;
    while (lead_s < s_len) {
      if (s[lead_s] > ' ')
        break;
      lead_s++;
    }
    memmove(s, &s[lead_s], 1 + (s_len - lead_s)); /* will copy null if zero len str */
    /* leading space gone.  now see to the trailing spaces */
    end_s = strlen(s);
    if (end_s > 0) {
      while (end_s-- > 0) {
        if (s[end_s] > ' ') { /* if non space .... */
          s[++end_s] = '\0';  /* put null on end of string */
          end_s = 0;          /* signal for end of loop */
        }
      }
    }
  } /* end if s_len */
}


void fill_defaults()
{
  int x;

  /* set some or all default (non-ini) values */
  if (out_bat_file_name[0] eq '\0') {
    strcpy(out_bat_file_name, "run_it.bat");
  }
  if (out_list_file_name[0] eq '\0') {
    strcpy(out_list_file_name, "list.txt");
  }
  if (leveled_monster_index eq 0) {
    for (x = 0; x neq 3; x++) {
      if ((leveled_monster[x] = malloc(40)) eq NULL) {
        fprintf(stderr, "fill defaults: memory failure\n");
        exit(DOOPS);
      }
    }
    strcpy(leveled_monster[0], "in_tomb_skele_lev+0");
    strcpy(leveled_monster[1], "in_tomb_bone_lev+0");
    strcpy(leveled_monster[2], "in_tomb_all_lev+0");
    leveled_monster_index = 3;
  }
  if (crate_index eq 0) {
    if ((crate[0] = malloc(20)) eq NULL) {
      fprintf(stderr, "fill defaults: memory failure\n");
      exit(DOOPS);
    }
    strcpy(crate[0], "crate_01_random_pos");
    crate_index = 1;
  }
  if (bag_index eq 0) {
    if ((bag[0] = malloc(16)) eq NULL) {
      fprintf(stderr, "fill defaults: memory failure\n");
      exit(DOOPS);
    }
    strcpy(bag[0], "com_sack_02_pos");
    bag_index = 1;
  }
  if (small_chest_index eq 0) {
    if ((small_chest[0] = malloc(22)) eq NULL) {
      fprintf(stderr, "fill defaults: memory failure\n");
      exit(DOOPS);
    }
    strcpy(small_chest[0], "chest_small_02_misc01");
    small_chest_index = 1;
  }
}


int parse_files(f)  /* section end is signaled by empty line (extra CR) */
FILE *f;
{
  char work_line[MAX_STR];
  int skip_rest, status;

  while (1) {
    if ((status = read_line(f, &work_line[0])) neq OKDOKEY)
      return(OKDOKEY);  /* assume normal EOF */
    trim_spaces(&work_line[0]);
    if (strlen(work_line) eq 0)  /* empty line (just a CR) */
      return(OKDOKEY);  /* end of section */
    skip_rest = 0;
    if (strncmp(&work_line[0], "include_esp_file", 16) eq 0) {
      skip_rest++;
      memmove(&work_line[0], &work_line[16], strlen(&work_line[0]) - 15);
      trim_spaces(&work_line[0]);
      strcpy(&include_esp_file_name[0], &work_line[0]);
    }
    if (skip_rest) continue;
    if (strncmp(work_line, "out_bat_file", 12) eq 0) {
      skip_rest++;
      memmove(work_line, &work_line[12], strlen(work_line) - 11);
      trim_spaces(&work_line[0]);
      strcpy(out_bat_file_name, work_line);
    }
    if (skip_rest) continue;
    if (strncmp(work_line, "out_list_file", 13) eq 0) {
      skip_rest++;
      memmove(work_line, &work_line[13], strlen(work_line) - 12);
      trim_spaces(&work_line[0]);
      strcpy(out_list_file_name, work_line);
    }
    if (skip_rest) continue;
    if (strncmp(work_line, "delete_temps", 12) eq 0) {
      del_files = 1;
      skip_rest++;
    }
    if (skip_rest) continue;
    fprintf(stderr, "parse_files: ignoring line \"%s\"\n", work_line);
  }
}


int parse_monsters(f) /* section end is signaled by empty line (extra CR) */
FILE *f;
{
  char work_line[MAX_STR];
  int status;

  while (1) {
    if ((status = read_line(f, &work_line[0])) neq OKDOKEY)
      return(OKDOKEY);
    trim_spaces(&work_line[0]);
    if (strlen(work_line) eq 0) /* end of monster section */
      return(OKDOKEY);
    if (leveled_monster_index < 20) {
      leveled_monster[leveled_monster_index] = malloc(1 + strlen(work_line));
      if (leveled_monster[leveled_monster_index] eq NULL) {
        fprintf(stderr, "memory failure in parse_monster\n");
        exit(DOOPS);
      }
      strcpy(leveled_monster[leveled_monster_index], work_line);
      leveled_monster_index++;
    }
  }
}


int parse_crate(f) /* section end is signaled by empty line (extra CR) */
FILE *f;
{
  char work_line[MAX_STR];
  int status;

  while (1) {
    if ((status = read_line(f, &work_line[0])) neq OKDOKEY)
      return(OKDOKEY);
    trim_spaces(&work_line[0]);
    if (strlen(work_line) eq 0) /* end of crate section */
      return(OKDOKEY);
    if (crate_index < 20) {
      crate[crate_index] = malloc(1 + strlen(work_line));
      if (crate[crate_index] eq NULL) {
        fprintf(stderr, "memory failure in parse_crate\n");
        exit(DOOPS);
      }
      strcpy(crate[crate_index], work_line);
      crate_index++;
    }
  }
}


int parse_bag(f) /* section end is signaled by empty line (extra CR) */
FILE *f;
{
  char work_line[MAX_STR];
  int status;

  while (1) {
    if ((status = read_line(f, &work_line[0])) neq OKDOKEY)
      return(OKDOKEY);
    trim_spaces(&work_line[0]);
    if (strlen(work_line) eq 0) /* end of bag section */
      return(OKDOKEY);
    if (bag_index < 20) {
      bag[bag_index] = malloc(1 + strlen(work_line));
      if (bag[bag_index] eq NULL) {
        fprintf(stderr, "memory failure in parse_bag\n");
        exit(DOOPS);
      }
      strcpy(bag[bag_index], work_line);
      bag_index++;
    }
  }
}


int parse_small_chest(f) /* section end is signaled by empty line (extra CR) */
FILE *f;
{
  char work_line[MAX_STR];
  int status;

  while (1) {
    if ((status = read_line(f, &work_line[0])) neq OKDOKEY)
      return(OKDOKEY);
    trim_spaces(&work_line[0]);
    if (strlen(work_line) eq 0) /* end of small chest section */
      return(OKDOKEY);
    if (small_chest_index < 20) {
      small_chest[small_chest_index] = malloc(1 + strlen(work_line));
      if (small_chest[small_chest_index] eq NULL) {
        fprintf(stderr, "memory failure in parse_small_chest\n");
        exit(DOOPS);
      }
      strcpy(small_chest[small_chest_index], work_line);
      small_chest_index++;
    }
  }
}


int parse_ini() /* complicated */
{
  char work_line[356];
  FILE *f;
  int status;

  if ((f = fopen("impgen.ini", "r")) eq NULL) {
    fprintf(stderr, "parse_ini: missing impgen.ini file - using defaults\n");
    return(DOOPS);  /* not a fatal error though */
  }
  /* at this point, there is a .ini file and it has been opened without
     error */
  while (1) {
    status = read_line(f, &work_line[0]);
    if (status eq EOF) {
      fclose(f);
      return(OKDOKEY);
    }
    trim_spaces(&work_line[0]);
    if (stricmp(work_line, "[files]") eq 0) {
      status = parse_files(f);
      if (status neq OKDOKEY) {
        fclose(f);
        return(status);
      }
    } else {
      if (stricmp(work_line, "[leveled_creature]") eq 0) {
        status = parse_monsters(f);
        if (status neq OKDOKEY) {
          fclose(f);
          return(OKDOKEY);
        }
      } else {
        if (stricmp(work_line, "[crate]") eq 0) {
          status = parse_crate(f);
          if (status neq OKDOKEY) {
            fclose(f);
            return(OKDOKEY);
          }
        } else {
          if (stricmp(work_line, "[bag]") eq 0) {
            status = parse_bag(f);
            if (status neq OKDOKEY) {
              fclose(f);
              return(OKDOKEY);
            }
          } else {
            if (stricmp(work_line, "[small_chest]") eq 0) {
              status = parse_small_chest(f);
              if (status neq OKDOKEY) {
                fclose(f);
                return(OKDOKEY);
              }
            } else {
              fprintf(stderr, "possible parse error - ignoring \"%s\"\n", work_line);
            }
          }
        }
      }
    }
  } /* end infinite while */
}


void test_parse() /* really just a display routine */
{
  int x;

  if (strlen(include_esp_file_name) eq 0) {
    printf("no include esp file\n");
  } else {
    printf("include_esp_file is: \"%s\"\n", include_esp_file_name);
  }
  printf("out_bat_file is: \"%s\"\n", out_bat_file_name);
  printf("out_list_file is: \"%s\"\n", out_list_file_name);

  printf("\nleveled_monsters\n");
  for (x = 0; x neq leveled_monster_index; x++) {
    printf("\"%s\"\n", leveled_monster[x]);
  }

  printf(
    "\ncrate\n"
  );

  for (x = 0; x neq crate_index; x++) {
    printf("\"%s\"\n", crate[x]);
  }
  printf("\nbag\n");
  for (x = 0; x neq bag_index; x++) {
    printf("\"%s\"\n", bag[x]);
  }
  printf("\nsmall_chest\n");
  for (x = 0; x neq small_chest_index; x++) {
    printf("\"%s\"\n", small_chest[x]);
  }

}


void usage()
{
  fprintf(stderr, "ImpGen by PCC\n\
Usage: impgen7 <base outfile name> <size> <base cell name> \
[-t] [-m] [-c] [-nt] [-l #] [-sl #]\n-t is no treasure, -m is no monsters, \
-c is no clutter (including lights), -nt is no twistyness, -l and a number \
specifies number of vertical levels, -sl and a number specifies the \
starting level.\n");
}


void usage_exit()
{
  usage();
  exit(DOOPS);
}


int one_hex(temp)
int temp;
{
  if (temp > 9) {
    return(temp + 'A' - 10);
  }
  return(temp + '0');
}


void m_str(n, the_str)  /* returns a 3 character string n must in in range -99 - 99 */
int n;
char *the_str;  /* must be 3 bytes in length (2 chars plus one null) */
{
  int temp;

  if (n < 0) {
    temp = 256 + n;
  } else {
    temp = n;
  }
  the_str[0] = one_hex(temp / 16);
  the_str[1] = one_hex(temp % 16);
  the_str[2] = '\0';
}


void cell_name(xx, yy, return_str)
int xx;
int yy;
char *return_str;
{
  if (num_levels eq 1) { /* single level */
    sprintf(return_str, "%s, (%d, %d)", base_cell_name, xx, yy);
  } else {  /* multi-level */
    sprintf(return_str, "%s, Level %d, (%d, %d)", base_cell_name, cur_level + lev_offset, xx, yy);
  }
}


int esp_grid(pov_magnitude)
int pov_magnitude;
{
  int rr;

  rr = (pov_magnitude * 256);
  return(rr);
}


void place_rug(xx, yy, zz)
int xx;
int yy;
int zz;
{
  char which_rug[40];
  int rug_rot, rug_ok;

  rug_ok = 1;
  if (no_clutter)
    rug_ok = 0;  /* rug is clutter - no rug */
  /* check to see if there is a trapddor down here */
  if (cur_level < num_levels)
    if ((xx eq new_dx) && (yy eq new_dy))
      rug_ok = 0;  /* no rug here */
  if (rug_ok) {
    if (rnd(3) eq 0) {
      strcpy(which_rug, "furn_de_rug_01");  /* bluish one */
    } else {
      strcpy(which_rug, "furn_de_rug_02");  /* non-bluish one */
    }
    rug_rot = 90 * rnd(4);
    if (rnd(10) > 6) {  /* perturb the rotation a little */
      rug_rot = (rug_rot - 4) + rnd(8);
    }
    /* write a rug ... */
    fprintf(
      _txtr_file,
      "reference \"%s\" 1 <0 0 %d> <%d %d %d>\n",
      which_rug,
      rug_rot,
      (-4 + rnd(8) + esp_grid(xx)),
      (-4 + rnd(8) - esp_grid(yy)),
      (-128 + esp_grid(zz))
    );
  }
}


void place_doorjam(rot, xx, yy, zz)
int rot;
int xx;
int yy;
int zz;
{
  int doorjam_rot;
  int doorjam_trans[3], door_trans[3];

  switch (rot) {
    case 0 : /* west wall doorway */
      doorjam_rot = 0;
      doorjam_trans[0] = -128;
      doorjam_trans[1] = 0;
      doorjam_trans[2] = 0;
      door_trans[0] = -130;
      door_trans[1] = 60;
      door_trans[2] = -32;
      break;
    case 90 :  /* north wall doorway */
      doorjam_rot = 90;
      doorjam_trans[0] = 0;
      doorjam_trans[1] = 128;
      doorjam_trans[2] = 0;
      door_trans[0] = 60;
      door_trans[1] = 130;
      door_trans[2] = -32;
      break;
    case 180 :  /* east wall doorway */
      doorjam_rot = 0;
      doorjam_trans[0] = 128;
      doorjam_trans[1] = 0;
      doorjam_trans[2] = 0;
      door_trans[0] = (-130 + 256);
      door_trans[1] = 60;
      door_trans[2] = -32;
      break;
    case 270 :  /* south wall doorway */
      doorjam_rot = 90;
      doorjam_trans[0] = 0;
      doorjam_trans[1] = -128;
      doorjam_trans[2] = 0;
      door_trans[0] = 60;
      door_trans[1] = (130 - 256);
      door_trans[2] = -32;
      break;
  }  /* end switch on rot */
  fprintf( // write a doorway
    _txtr_file,
    "reference \"in_impsmall_doorjam_01\" 1 <0 0 %d> <%d %d %d>\n",
    doorjam_rot,
    doorjam_trans[0] + esp_grid(xx),
    doorjam_trans[1] + 0 - esp_grid(yy),
    doorjam_trans[2] + esp_grid(zz)
  );
}


void place_door(rot, xx, yy, zz)
int rot;
int xx;
int yy;
int zz;
{
  int doorjam_rot;
  int door_trans[3];

  switch (rot) {
    case 0 :  /* west wall doorway */
      doorjam_rot = 0;
      door_trans[0] = -130;
      door_trans[1] = 60;
      door_trans[2] = -32;
      break;
    case 90 :  /* north wall doorway */
      doorjam_rot = 90;
      door_trans[0] = 60;
      door_trans[1] = 130;
      door_trans[2] = -32;
      break;
    case 180 :  /* east wall doorway */
      doorjam_rot = 0;
      door_trans[0] = -130 + 256;
      door_trans[1] = 60;
      door_trans[2] = -32;
      break;
    case 270 :  /* south wall doorway */
      doorjam_rot = 90;
      door_trans[0] = 60;
      door_trans[1] = 130 - 256;
      door_trans[2] = -32;
      break;
  }  /* end switch on rot */
  fprintf(
    _txtr_file,
    "reference \"in_impsmall_door_01\" 1 <0 0 %d> <%d %d %d>\n",
    doorjam_rot,
    door_trans[0] + esp_grid(xx),
    door_trans[1] + 0 - esp_grid(yy),
    door_trans[2] + esp_grid(zz)
  );
}


void place_load_door(rot, xx, yy, zz, m_x, m_y, dungeon_size)
int rot;
int xx;
int yy;
int zz;
int m_x;
int m_y;
int dungeon_size;
{
  int teleport_ok = 1;  /* anything not zero is true */
  int doorjam_rot;
  int door_trans[3];
  int dest_coords[3];
  int dest_rot;
  int dest_cell[2];  /* only a 2-vector - no Z needed */
  char temp_cell_name[MAX_STR];


  switch (rot) {
    case 0 :  /* west wall doorway */
      doorjam_rot = 0;
      door_trans[0] = -128;
      door_trans[1] = 0;
      door_trans[2] = 0;
      /* destination pink arrow */
      dest_coords[0] = esp_grid(maze_size);
      dest_coords[1] = 0 - esp_grid( (maze_size / 2) );
      dest_coords[2] = 0;
      dest_rot = 270;  /* point west */
      dest_cell[0] = ROAi(m_x) - 1;
      dest_cell[1] = m_y;
      /* is there a neighboring west cell to teleport to? */
      if (ROAi(m_x) <= (1 - ROAi(dungeon_size))) {
        teleport_ok = 0;  /* set to false */
      }
      break;
    case 90 :  /* north wall doorway */
      doorjam_rot = 90;
      door_trans[0] = 0;
      door_trans[1] = 128;
      door_trans[2] = 0;
      /* destination pink arrow */
      dest_coords[0] = esp_grid( (maze_size / 2) );
      dest_coords[1] = 0 - esp_grid(maze_size);
      dest_coords[2] = 0;
      dest_rot = 0; // point north
      dest_cell[0] = m_x;
      dest_cell[1] = ROAi(m_y) - 1;
      /* is there a neighboring north cell to teleport to? */
      if (ROAi(m_y) <= (1 - ROAi(dungeon_size))) {
        teleport_ok = 0;  /* set to false */
      }
      break;
    case 180 :  /* east wall doorway */
      doorjam_rot = 180;
      door_trans[0] = 128;
      door_trans[1] = 0;
      door_trans[2] = 0;
      /* destination pink arrow */
      dest_coords[0] = esp_grid(-1);
      dest_coords[1] = 0 - esp_grid( (maze_size / 2) );
      dest_coords[2] = 0;
      dest_rot = 90;   /* point east */
      dest_cell[0] = ROAi(m_x) + 1;
      dest_cell[1] = m_y;
      /* is there a neighboring east cell to teleport to? */
      if (ROAi(m_x) >= (-1 + ROAi(dungeon_size))) {
        teleport_ok = 0; /* set to false */
      }
      break;
    case 270 :  /* south wall doorway */
      doorjam_rot = 270;
      door_trans[0] = 0;
      door_trans[1] = -128;
      door_trans[2] = 0;
      /* destination pink arrow */
      dest_coords[0] = esp_grid( (maze_size / 2) );
      dest_coords[1] = 0 - esp_grid(-1);
      dest_coords[2] = 0;
      dest_rot = 180;   /* point south */
      dest_cell[0] = m_x;
      dest_cell[1] = ROAi(m_y) + 1;
      /* is there a neighboring south cell to teleport to? */
      if (ROAi(m_y) >= (-1 + ROAi(dungeon_size))) {
        teleport_ok = 0;  /* set to false */
      }
      break;
  } // end switch on rot

  if (teleport_ok neq 0) {  /* if true, teleporting door */
    cell_name(dest_cell[0], dest_cell[1], &temp_cell_name[0]);
    fprintf(
      _txtr_file,
      "tel_door_ref \"in_impsmall_loaddoor_01\" 1 <0 0 %d> <%d %d %d> <0 0 %d> <%d %d %d> \"%s\"\n",
      doorjam_rot,
      door_trans[0] + esp_grid(xx),
      door_trans[1] + 0 - esp_grid(yy),
      door_trans[2] + esp_grid(zz),
      /* now for teleport data */
      dest_rot,
      dest_coords[0], dest_coords[1], dest_coords[2],
      temp_cell_name  /* dest cell */
    );
  } else { /* non-teleporting door - to be manually hooked up later? */
    fprintf(
      _txtr_file,
      "reference \"in_impsmall_loaddoor_01\" 1 <0 0 %d> <%d %d %d>\n",
      doorjam_rot,
      door_trans[0] + esp_grid(xx),
      door_trans[1] + 0 - esp_grid(yy),
      door_trans[2] + esp_grid(zz)
    );
  }
}


void place_tapestry(rot, xx, yy, zz)
int rot;
int xx;
int yy;
int zz;
{
  char which_tapestry[40];
  int coords[4];

  if (no_clutter eq 0) {
    switch (rnd(7)) {
      case (0) : strcpy(which_tapestry, "furn_com_tapestry_01"); break;
      case (1) : strcpy(which_tapestry, "furn_com_tapestry_02"); break;
      case (2) : strcpy(which_tapestry, "furn_com_tapestry_03"); break;
      case (3) : strcpy(which_tapestry, "furn_com_tapestry_05"); break;
      default : /* imperial */
        strcpy(which_tapestry, "furn_com_tapestry_04");  /* imperial */
    }  /* end switch */

    coords[0] = 180;  /* rotate */
    coords[1] = 0;
    coords[2] = 97;
    coords[3] = 80;
    RTRT(&coords[0], (180 + rot) mod 360, esp_grid(xx), 0 - esp_grid(yy), esp_grid(zz));

    fprintf(
      _txtr_file,
      "reference \"%s\" 1 <0 0 %d> <%d %d %d>\n",
      which_tapestry,
      coords[0],
      coords[1],
      coords[2],
      coords[3]
    );
  }
}


void place_monster(xx, yy, zz)
int xx;
int yy;
int zz;
{
  int no_monster_allowed = 0;  /* starts out as false */
  char which_monster[40];

  if (xx <= 0) {
    no_monster_allowed = 1;  /* true */
  }
  if (xx >= (maze_size - 1)) {
    no_monster_allowed = 1;  /* true */
  }
  if (yy <= 0) {
    no_monster_allowed = 1;  /* true */
  }
  if (yy >= (maze_size - 1)) {
    no_monster_allowed = 1;  /* true */
  }
  if (no_monsters)  /* command line option for no monsters */
    no_monster_allowed = 1;  /* so no monsters */

  /* check for trapdoors */
  if (num_levels > 1) {  /* only check on a multi-level dungeon */
    if (cur_level > 1) { /* check for door to/from level above */
      if (last_dx eq xx) {
        if (last_dy eq yy ) {
          no_monster_allowed = 1;
        }
      }
    }
    if (cur_level < num_levels) { /* check for door to/from level below */
      if (new_dx eq xx) {
        if (new_dy eq yy ) {
          no_monster_allowed = 1;
        }
      }
    }
  }

  if (no_monster_allowed eq 0) {  /* if monster is allowed in this square */
    /* this only works with a leveled creature */
    fprintf(
      _txtr_file, "reference \"%s\" 1 <0 0 %d> <%d %d %d>\n",
      leveled_monster[rnd(leveled_monster_index)],
      rnd(360),
      esp_grid(xx),
      0 - esp_grid(yy),
      -112 + esp_grid(zz)
    );
  }
}


void place_4way(esp_x, esp_y, esp_z)
int esp_x;
int esp_y;
int esp_z;
{
  fprintf(
    _txtr_file,
    "reference \"in_impsmall_4way_01\" 1 <0 0 0> <%d %d %d>\n",
    esp_grid(esp_x), (0 - esp_grid(esp_y)), esp_grid(esp_z)
  );
  if (rnd(12) eq 3) {
    place_monster(esp_x, esp_y, esp_z);
  }
}


void place_straight(rot, esp_x, esp_y, esp_z)
int rot;
int esp_x;
int esp_y;
int esp_z;
{
  char which_hall[40];
  int vert_offset;

  switch(rnd(3)) {
    case (0) :
      strcpy(which_hall, "in_impsmall_hall_02"); /* different style - lower poly count */
      vert_offset = -128;  /* but out of line with the usual style! */
      break;
    default :
      strcpy(which_hall, "in_impsmall_hall_01");  /* usual style */
      vert_offset = 0;   /* this one is in line */
      if (rnd(2) eq 1) {
        place_tapestry( ((rot + 90) % 360), esp_x, esp_y, esp_z);
      }
      if (rnd(2) eq 1) {
        place_tapestry( ((rot + 270) % 360), esp_x, esp_y, esp_z);
      }
  } /* end switch */
  fprintf(
    _txtr_file,
    "reference \"%s\" 1 <0 0 %d> <%d %d %d>\n",
    which_hall,
    rot,
    esp_grid(esp_x),
    (0 - esp_grid(esp_y)),
    (esp_grid(esp_z) + vert_offset)
  );
  if (rnd(12) eq 3) {
    place_monster(esp_x, esp_y, esp_z);
  }
}


void place_corner(rot, esp_x, esp_y, esp_z)
int rot;
int esp_x;
int esp_y;
int esp_z;
{
  int use_light, light_rot, light_z;
  int light_trans[3];
  char which_light[40];

  fprintf(
    _txtr_file,
    "reference \"in_impsmall_corner_01\" 1 <0 0 %d> <%d %d %d>\n",
    rot,
    esp_grid(esp_x), (0 - esp_grid(esp_y)), esp_grid(esp_z)
  );
  /* put some random clutter in */
  use_light = rnd(2);
  if (no_clutter)  /* command line switch for no lights */
    use_light = 1;
  if (use_light eq 0) {
    /* light in corner */
    switch (rnd(5)) {
      case 0 : strcpy(which_light, "light_de_lamp_01_177");  light_z = 0; break;
      case 1 : strcpy(which_light, "light_de_lamp_06_256");  light_z = 8; break;
      case 2 : strcpy(which_light, "light_de_lamp_02_blue_128");  light_z = 3; break;
      case 3 : strcpy(which_light, "light_de_lamp_03_256");  light_z = 5; break;
      default : strcpy(which_light, "light_de_lamp_04_128");  light_z = 6;
    }  /* end switch */
    light_rot = 43 + rnd(4);
    switch (rot) {
      case 0 :  /* north west corner */
        light_trans[0] = -76 - rnd(4);
        light_trans[1] = 76 + rnd(4);
        light_trans[2] = light_z;
        break;
      case 90 :  /* north east corner */
        light_trans[0] = 76 - rnd(4);
        light_trans[1] = 76 + rnd(4);
        light_trans[2] = light_z;
        break;
      case 180 :  /* south east corner */
        light_trans[0] = 76 + rnd(4);
        light_trans[1] = -76 - rnd(4);
        light_trans[2] = light_z;
        break;
      default :
        light_trans[0] = -76 - rnd(4);
        light_trans[1] = -76 - rnd(4);
        light_trans[2] = light_z;
    } /* end switch */
    fprintf(
      _txtr_file,
      "reference \"%s\" 1 <0 0 %d> <%d %d %d>\n",
      which_light,
      light_rot,
      light_trans[0] + esp_grid(esp_x),
      light_trans[1] + 0 - esp_grid(esp_y),
      light_trans[2] + esp_grid(esp_z)
    );
  }
  if (rnd(12) eq 3) {
    place_monster(esp_x, esp_y, esp_z);
  }
  if (rnd(10) > 4) { place_tapestry( ((rot + 90) % 360), esp_x, esp_y, esp_z); }
  if (rnd(10) > 4) { place_tapestry( ((rot + 180) % 360), esp_x, esp_y, esp_z); }
}


void place_3way(rot, esp_x, esp_y, esp_z)
int rot;
int esp_x;
int esp_y;
int esp_z;
{
  fprintf(
    _txtr_file,
    "reference \"in_impsmall_3way_01\" 1 <0 0 %d> <%d %d %d>\n",
    rot,
    esp_grid(esp_x), 0 - esp_grid(esp_y), esp_grid(esp_z)
  );
  if (rnd(12) eq 3) {
    place_monster(esp_x, esp_y, esp_z);
  }
}


void place_cap(rot, esp_x, esp_y, esp_z)
int rot;
int esp_x;
int esp_y;
int esp_z;
{
  fprintf(
    _txtr_file,
    "reference \"in_impsmall_endcap_01\" 1 <0 0 %d> <%d %d %d>\n",
    rot,
    esp_grid(esp_x), 0 - esp_grid(esp_y), esp_grid(esp_z)
  );
  if (rnd(12) eq 3) {
    place_monster(esp_x, esp_y, esp_z);
  }
  if (rnd(2) eq 1) { place_tapestry( ((rot + 90) % 360), esp_x, esp_y, esp_z); }
  if (rnd(2) eq 1) { place_tapestry( ((rot + 180) % 360), esp_x, esp_y, esp_z); }
  if (rnd(2) eq 1) { place_tapestry( ((rot + 270) % 360), esp_x, esp_y, esp_z); }
}


void place_corridor(xx, yy, zz)
int xx;
int yy;
int zz;
{
  /* 2-way N_S, 2-way E-W, corner (4 rots), 3way (4 rots), 4-way, cap
     remember our NSWE code?  well, just use that in a case statement */
  switch(maze[xx][yy]) {
    case 12 :  /* N-S straight corridor */
      place_straight(0, xx, yy, zz);
      break;
    case 3 :  /* W-E straight corridor */
      place_straight(90, xx, yy, zz);
      break;
    case 5 :  /* corner South and East */
      place_corner(0, xx, yy, zz);
      break;
    case 6 :  /* corner West and South */
      place_corner(90, xx, yy, zz);
      break;
    case 10 :  /* corner West and North */
      place_corner(180, xx, yy, zz);
      break;
    case 9 :  /* corner North and East */
      place_corner(270, xx, yy, zz);
      break;
    case 13 :  /* 3way, closed on West */
      place_3way(0, xx, yy, zz);
      break;
    case 7 :  /* 3way, closed on North */
      place_3way(90, xx, yy, zz);
      break;
    case 14 :  /* 3way, closed on East */
      place_3way(180, xx, yy, zz);
      break;
    case 11 :  /* 3way, closed on South */
      place_3way(270, xx, yy, zz);
      break;
    case 4 :  /* endcap (nub) open on South */
      place_cap(0, xx, yy, zz);
      break;
    case (2) :  /* endcap (nub) open on West */
      place_cap(90, xx, yy, zz);
      break;
    case (8) :  /* endcap (nub) open on North */
      place_cap(180, xx, yy, zz);
      break;
    case (1) :  /* endcap (nub) open on East */
      place_cap(270, xx, yy, zz);
      break;
    default :  /* default case */
      place_4way(xx, yy, zz);
  }  /* end switch */
  if (rnd(10) > 8) {
    place_rug(xx, yy, zz);
  }
}


/* end of corridor crap.  now room crap */


void place_room_entry(rot, esp_x, esp_y, esp_z)
int rot;
int esp_x;
int esp_y;
int esp_z;
{
  fprintf(
    _txtr_file,
    "reference \"in_impsmall_r_entr_01\" 1 <0 0 %d> <%d %d %d>\n",
    rot,
    esp_grid(esp_x), 0 - esp_grid(esp_y), esp_grid(esp_z)
  );
  /* place doorjam */
  place_doorjam(rot, esp_x, esp_y, esp_z);
  if (rnd(3) > 0) { /* write a door */
    place_door(rot, esp_x, esp_y, esp_z);
  }
  if (rnd(10) > 7) { place_rug(esp_x, esp_y, esp_z); }
}


void place_room_corner(rot, esp_x, esp_y, esp_z)
int rot;
int esp_x;
int esp_y;
int esp_z;
{
  fprintf(
    _txtr_file,
    "reference \"in_impsmall_r_corner_01\" 1 <0 0 %d> <%d %d %d>\n",
    rot,
    esp_grid(esp_x), 0 - esp_grid(esp_y), esp_grid(esp_z)
  );
  if (rnd(10) > 4) place_tapestry( ((rot + 90) % 360), esp_x, esp_y, esp_z);
  if (rnd(10) > 4) place_tapestry( ((rot + 180) % 360), esp_x, esp_y, esp_z);
  if (rnd(10) > 7) place_rug(esp_x, esp_y, esp_z);
}


void place_room_concave_corner(rot, esp_x, esp_y, esp_z) /* concave corner */
int rot;
int esp_x;
int esp_y;
int esp_z;
{
  fprintf(
    _txtr_file,
    "reference \"in_impsmall_r_corner_02\" 1 <0 0 %d> <%d %d %d>\n",
    rot,
    esp_grid(esp_x), 0 - esp_grid(esp_y), esp_grid(esp_z)
  );
  if (rnd(10) > 7) { place_rug(esp_x, esp_y, esp_z); }
}


void place_room_center(esp_x, esp_y, esp_z)
int esp_x;
int esp_y;
int esp_z;
{
  fprintf(
    _txtr_file,
    "reference \"in_impsmall_r_center_01\" 1 <0 0 0> <%d %d %d>\n",
    esp_grid(esp_x), 0 - esp_grid(esp_y), esp_grid(esp_z)
  );
  if (rnd(10) > 7) { place_rug(esp_x, esp_y, esp_z); }
}


void place_room_side(rot, esp_x, esp_y, esp_z)
int rot;
int esp_x;
int esp_y;
int esp_z;
{
  fprintf(
    _txtr_file,
    "reference \"in_impsmall_r_3way_01\" 1 <0 0 %d> <%d %d %d>\n",
    rot,
    esp_grid(esp_x), 0 - esp_grid(esp_y), esp_grid(esp_z)
  );
  if (rnd(10) > 4) place_tapestry( ((rot + 90) % 360), esp_x, esp_y, esp_z);
  if (rnd(10) > 7) place_rug(esp_x, esp_y, esp_z);
}


void place_room_wonky1(esp_x, esp_y, esp_z)
int esp_x;
int esp_y;
int esp_z;
{
  fprintf(
    _txtr_file,
    "reference \"in_impsmall_r_center_01\" 1 <0 0 0> <%d %d %d>\n",
    esp_grid(esp_x), 0 - esp_grid(esp_y), esp_grid(esp_z)
  );
  fprintf( /* north west pillar */
    _txtr_file,
    "reference \"in_impsmall_r_pillar_01\" 1 <0 0 0> <%d %d %d>\n",
    esp_grid(esp_x) - 125, 0 - esp_grid(esp_y) + 125, esp_grid(esp_z)
  );
  fprintf( /* south east pillar */
    _txtr_file,
    "reference \"in_impsmall_r_pillar_01\" 1 <0 0 0> <%d %d %d>\n",
    esp_grid(esp_x) + 125, 0 - esp_grid(esp_y) - 125, esp_grid(esp_z)
  );
  if (rnd(10) > 7) place_rug(esp_x, esp_y, esp_z);
}


void place_room_wonky2(esp_x, esp_y, esp_z)
int esp_x;
int esp_y;
int esp_z;
{
  fprintf(
    _txtr_file,
    "reference \"in_impsmall_r_center_01\" 1 <0 0 0> <%d %d %d>\n",
    esp_grid(esp_x), 0 - esp_grid(esp_y), esp_grid(esp_z)
  );
  fprintf(  /* north east pillar */
    _txtr_file,
    "reference \"in_impsmall_r_pillar_01\" 1 <0 0 0> <%d %d %d>\n",
    esp_grid(esp_x) + 125, 0 - esp_grid(esp_y) + 125, esp_grid(esp_z)
  );
  fprintf( /* south west pillar */
    _txtr_file,
    "reference \"in_impsmall_r_pillar_01\" 1 <0 0 0> <%d %d %d>\n",
    esp_grid(esp_x) - 125, 0 - esp_grid(esp_y) - 125, esp_grid(esp_z)
  );
  if (rnd(10) > 7) place_rug(esp_x, esp_y, esp_z);
}


void place_room(xx, yy, zz)
int xx;
int yy;
int zz;
{
  int cc, is_entry, NO_clutter, room_flags, put_crate, put_monster;
  char which_candle[64];

  /* entrance (3 types), center, corner, concave corner, room side
     a special case is if two rooms touch corner-to-corner.  In that
     case we either wall them apart, XOR we use two odd-entryways to connect
     them.  Try to use teh first method, cause some tile-sets don't
     have corner entryway peices (imperial and vivic has them though)
    
     First check if it is a room entrance */
  cc = 0;
  is_entry = 0; /* false */
  while (cc < room_doors_index) {
    if (xx eq room_doors_x[cc]) {
      if (yy eq room_doors_y[cc]) {
        /* at this point, there could be a duplicate door
           that should not be rendered to the ESP but will still be
           in the doors array.  A way to verify against this is
           to see if that square has been flagged as a room door (2)
           or just a room (1) */
        if (rooms[xx][yy] eq 2) { /* 2 is code for rooom door */
          is_entry = 1;  /* yes, this is a room entrance - set true */
          cc = room_doors_index; /* signal early exit from loop */
        }
      }
    }
    cc++;
  }
  if (is_entry neq 0) {  /* if is_entry is true */
    /* Since it is a room entrance, find where the door goes!  What makes
       this easier is that the square will be surrounded on three sides by
       room peices.  All we must do is find which side is adjacent to
       a non room peice.  Ignore diagonals. */
    if (rooms[xx][(yy - 1)] eq 0) {  /* door is to the north */
      place_room_entry(90, xx, yy, 0);
    } else {
      if (rooms[xx][(yy + 1)] eq 0) {  /* door is to the south */
        place_room_entry(270, xx, yy, 0);
      } else {
        if (rooms[(xx - 1)][yy] eq 0) {  /* door is to the west */
          place_room_entry(0, xx, yy, 0);
        } else {  /* door is to the east */
          place_room_entry(180, xx, yy, 0);
        }
      }
    }
  } else {  /* a room peice but not a room entrance */
    /* this is very very complicated.  Consider two rooms touching
       corner to corner, or corner to corner slid over one on either the x
       or y.  Eeep!
       Ok, so we do this a clumsy method that requires less thinking but
       more run time.  We take a look at all 8 cells adjacent to the room
       and use it as a bit pattern to construct one number.
       Then we take that number and use it in a case statement.  Ugly
       but workable */
    NO_clutter = 0; /* set to false */
    room_flags = 0;
    if (rooms[(xx - 1)][(yy - 1)] > 0) {
      room_flags = room_flags + 128;
    }
    if (rooms[xx][(yy - 1)] > 0) {
      room_flags = room_flags + 64;
    }
    if (rooms[(xx + 1)][(yy - 1)] > 0) {
      room_flags = room_flags + 32;
    }
    if (rooms[(xx + 1)][yy] > 0) {
      room_flags = room_flags + 16;
    }
    if (rooms[(xx + 1)][(yy + 1)] > 0) {
      room_flags = room_flags + 8;
    }
    if (rooms[xx][(yy + 1)] > 0) {
      room_flags = room_flags + 4;
    }
    if (rooms[(xx - 1)][(yy + 1)] > 0) {
      room_flags = room_flags + 2;
    }
    if (rooms[(xx - 1)][yy] > 0) {
      room_flags = room_flags + 1;
    }
    switch (room_flags) {
      /* cases for north west corner */
      case ((16 + 8 + 4)) :
      case ((16 + 8 + 4 + 32)) :
      case ((16 + 8 + 4 + 2)) :
      case ((16 + 8 + 4 + 32 + 2)) :
      case ((16 + 8 + 4 + 128)) :
      case ((16 + 8 + 4 + 128 + 1)) :
      case ((16 + 8 + 4 + 128 + 64)) :
      case ((16 + 8 + 4 + 128 + 32 + 1)) :
      case ((16 + 8 + 4 + 2 + 128 + 64)) :
      case ((16 + 8 + 4 + 128 + 32)) :
      case ((16 + 8 + 4 + 128 + 2)) :
        place_room_corner(0, xx, yy, 0);
        break;
      /* cases for north east corner */
      case ((1 + 2 + 4)) :
      case ((1 + 2 + 4 + 8)):
      case ((1 + 2 + 4 + 128)) :
      case ((1 + 2 + 4 + 8 + 128)) :
      case ((1 + 2 + 4 + 32)) :
      case ((1 + 2 + 4 + 32 + 64)) :
      case ((1 + 2 + 4 + 32 + 16)) :
      case ((1 + 2 + 4 + 128 + 32 + 16)) :
      case ((1 + 2 + 4 + 64 + 32 + 8)) :
      case ((1 + 2 + 4 + 128 + 32)) :
      case ((1 + 2 + 4 + 32 + 8)) :
        place_room_corner(90, xx, yy, 0);
        break;
      /* cases for south east corner */
      case ((1 + 128 + 64)) :
      case ((1 + 128 + 64 + 32)) :
      case ((1 + 128 + 64 + 2)) :
      case ((1 + 128 + 64 + 32 + 2)) :
      case ((1 + 128 + 64 + 8)) :
      case ((1 + 128 + 64 + 8 + 16)) :
      case ((1 + 128 + 64 + 8 + 4)) :
      case ((1 + 128 + 64 + 32 + 8 + 4)) :
      case ((1 + 128 + 64 + 16 + 8 + 2)) :
      case ((1 + 128 + 64 + 32 + 8)) :
      case ((1 + 128 + 64 + 8 + 2)) :
        place_room_corner(180, xx, yy, 0);
        break;
      /* cases for south west corner */
      case ((64 + 32 + 16)) :
      case ((64 + 32 + 16 + 128)) :
      case ((64 + 32 + 16 + 8)) :
      case ((64 + 32 + 16 + 128 + 8)) :
      case ((64 + 32 + 16 + 2)) :
      case ((64 + 32 + 16 + 2 + 1)) :
      case ((64 + 32 + 16 + 2 + 4)) :
      case ((64 + 32 + 16 + 8 + 2 + 1)) :
      case ((64 + 32 + 16 + 128 + 4 + 2)) :
      case ((64 + 32 + 16 + 8 + 2)) :
      case ((64 + 32 + 16 + 128 + 2)) :
        place_room_corner(270, xx, yy, 0);
        break;
      /* case for room north side */
      case ((1 + 2 + 4 + 8 + 16)) :
      case ((1 + 2 + 4 + 8 + 16 + 128)) :
      case ((1 + 2 + 4 + 8 + 16 + 32)) :
      case ((1 + 2 + 4 + 8 + 16 + 128 + 32)) :
        place_room_side(90, xx, yy, 0);
        break;
      /* case for room east side */
      case ((64 + 128 + 1 + 2 + 4)) :
      case ((64 + 128 + 1 + 2 + 4 + 32)) :
      case ((64 + 128 + 1 + 2 + 4 + 8)) :
      case ((64 + 128 + 1 + 2 + 4 + 8 + 32)) :
        place_room_side(180, xx, yy, 0);
        break;
      /* case for room south side */
      case ((1 + 128 + 64 + 32 + 16)) :
      case ((1 + 128 + 64 + 32 + 16 + 2)) :
      case ((1 + 128 + 64 + 32 + 16 + 8)) :
      case ((1 + 128 + 64 + 32 + 16 + 8 + 2)) :
        place_room_side(270, xx, yy, 0);
        break;
      /* case for room west side */
      case ((64 + 32 + 16 + 8 + 4)) :
      case ((64 + 32 + 16 + 8 + 4 + 128)) :
      case ((64 + 32 + 16 + 8 + 4 + 2)) :
      case ((64 + 32 + 16 + 8 + 4 + 2 + 128)) :
        place_room_side(0, xx, yy, 0);
        break;
      /* case for concave corner, north east */
      case ((1 + 2 + 4 + 8 + 16 + 64 + 128)) :
        place_room_concave_corner(90, xx, yy, 0);
        break;
      /* case for concave corner, south east */
      case ((1 + 2 + 4 + 16 + 32 + 64 + 128)) :
        place_room_concave_corner(180, xx, yy, 0);
        break;
      /* case for concave corner, south west */
      case ((1 + 4 + 8 + 16 + 32 + 64 + 128)) :
        place_room_concave_corner(270, xx, yy, 0);
        break;
      /* case for concave corner, north west */
      case ((1 + 2 + 4 + 8 + 16 + 32 + 64)) :
        place_room_concave_corner(0, xx, yy, 0);
        break;
      /* special dual corner peice - hard to describe.  will need it when
         two rooms overlap corner-on-corner
         there is no peice for this in the imperial tile set
         we fake it using a room center and two pillars */
      case ((64 + 32 + 16 + 1 + 2 + 4)) :
        place_room_wonky1(xx, yy, zz);
        NO_clutter = 1; /* set no clutter to be true */
        break;
      case ((128 + 64 + 1 + 4 + 8 + 16)) :
        place_room_wonky2(xx, yy, zz);
        NO_clutter = 1;  /* set no clutter to be true */
        break;
      default : /* default case, just put a room center in */
        place_room_center(xx, yy, zz);
    } /* end switch */

    /* check for trap door down (up is OK) before putting clutter */
    if (cur_level < num_levels)
      if ((xx eq new_dx) && (yy eq new_dy))
        NO_clutter = 1;  /* set no clutter to be true */

    /* some room clutter that can be placed anywhere but an entrance */
    if (NO_clutter eq 0) { /* if no clutter is still false */
      put_crate = rnd(4);
    } else {
      put_crate = 1; /* no crate  (0 would be yes crate) */
    }
    if (no_treasure)
      put_crate = 1; /* no crate  (0 would be yes crate) */
    put_monster = rnd(5);
    if ((last_dx eq xx) && (last_dy eq yy)) {
      put_monster = 4;  /* no monster under trap door */
    }
    if ((new_dx eq xx) && (new_dy eq yy)) {
      put_monster = 4;  /* no monster over trap door */
    }
    if (put_crate eq 0) {  /* 0 is yes crate */
      put_monster = 4;     /* anti-collision */
      fprintf(  /* crate */
        _txtr_file,
        "reference \"%s\" 1 <0 0 %d> <%d %d %d>\n",
        crate[rnd(crate_index)],
        rnd(360),
        -10 + rnd(20) + esp_grid(xx),
        -10 + rnd(20) + 0 - esp_grid(yy),
        -96 + esp_grid(zz)
      );
      if (rnd(3) eq 2) {  /* put something on top of the crate? */
        switch (rnd(4)) {
          case 0 :  /* another crate */
            fprintf(
              _txtr_file,
              "reference \"%s\" 1 <0 0 %d> <%d %d %d>\n",
              crate[rnd(crate_index)],
              rnd(360),
              -5 + rnd(10) + esp_grid(xx),
              -5 + rnd(10) + 0 - esp_grid(yy),
              -32 + esp_grid(zz)
            );
            break;
          case 1 :  /* little money chest */
            fprintf(
              _txtr_file,
              "reference \"%s\" 0.8 <0 0 %d> <%d %d %d>\n",
              small_chest[rnd(small_chest_index)],
              rnd(360),
              esp_grid(xx),
              0 - esp_grid(yy),
              -60 + esp_grid(zz)
            );
            break;
          case 2 :  /* 3 squat candles */
            switch (rnd(5)) {
              case 0 : strcpy(which_candle, "light_com_candle_04");  break; /* blue */
              case 1 : strcpy(which_candle, "light_com_candle_06");  break; /* red */
              case 2 : strcpy(which_candle, "light_com_candle_07");  break; /* off white */
              case 3 : strcpy(which_candle, "light_com_candle_11");  break; /* blue, blue light */
              default : strcpy(which_candle, "light_com_candle_13"); /* red, red light */
            }
            fprintf(
              _txtr_file,
              "reference \"%s\" 0.8 <0 0 %d> <%d %d %d>\n",
              which_candle,
              rnd(360),
              esp_grid(xx),
              0 - esp_grid(yy),
              -47 + esp_grid(zz)
            );
            break;
          default : /* bag */
            fprintf(
              _txtr_file,
              "reference \"%s\" 0.8 <0 0 %d> <%d %d %d>\n",
              bag[rnd(bag_index)],
              rnd(360),
              -5 + rnd(10) + esp_grid(xx),
              -5 + rnd(10) + 0 - esp_grid(yy),
              -46 + esp_grid(zz)
            );
        }  /* end switch */
      }
    }
    if (put_monster eq 0) {
      place_monster(xx, yy, zz);
    }
  }
  /* maybe place a (candle) chandelier */
  if (no_clutter eq 0) {
    if ((xx neq last_dx) || (yy neq last_dy)) { /* no light from a ceiling trap door */
      if (rnd(4) eq 1) {
        fprintf(
          _txtr_file,
          "reference \"light_com_chandelier_01\" 1 <0 0 %d> <%d %d %d>\n",
          rnd(360),
          -2 + rnd(4) + esp_grid(xx),
          -2 + rnd(4) + 0 - esp_grid(yy),
          81 + esp_grid(zz)
        );
      }
    }
  }
}


/* end of esp output macros.  the rest is the maze generator */


int check_door(door_loc_x, door_loc_y) /* check if a room door is possible at this <x, y> */
int door_loc_x;
int door_loc_y;
{
  int return_code, a_count, yy, xx;

  /* current coords must be in a room */
  if (rooms[door_loc_x][door_loc_y] eq 0) {  /* not a room */
    return_code = 0;   /* can't put a room doorway here! */
  } else {  /* current coords are in a room */
    /* one non-diagonal square must be non-room - the other three must be room
       this must not be an anadvertant concave corner formed by union of
       two overlapping rooms
       Near as I can tell, the fastest way to test that all is to count the number
       or foom squares out of the 8 adjacent cells.  The amount of room cells
       must be six. */
    a_count = 0;
    yy = 0;
    while (yy < 3) {
      xx = 0;
      while (xx < 3) {
        if (rooms[(door_loc_x - 1 + xx)][(door_loc_y - 1 + yy)] eq 1) {  /* is a room */
          a_count++;
        }
        xx++;
      }
      yy++;
    }
    if (a_count eq 6) {  /* can put a door here MAYBE */
      return_code = 1;
      if (rooms[(door_loc_x - 1)][(door_loc_y)] eq 2) {  /* non diagonal door neighbor? */
        return_code = 0;  /* nope - no door here */
      }
      if (rooms[(door_loc_x + 1)][(door_loc_y)] eq 2) {  /* non diagonal door neighbor? */
        return_code = 0;  /* nope - no door here */
      }
      if (rooms[(door_loc_x)][(door_loc_y - 1)] eq 2) {  /* non diagonal door neighbor? */
        return_code = 0;  /* nope - no door here */
      }
      if (rooms[(door_loc_x)][(door_loc_y + 1)] eq 2) {  /* non diagonal door neighbor? */
        return_code = 0;  /* nope - no door here */
      }
    } else {
      return_code = 0;  /* nope - no door here */
    }
  }
  return(return_code);
}


void find_doors(try_doors_x, try_doors_y) /*array of 4 2-vectors that are indexed into maze map */
int *try_doors_x;
int *try_doors_y;
{
  int try2_x[4];
  int try2_y[4];
  int try2_index = 0;
  int cc, temp, remove_this, try3;

  cc = 0;
  while (cc < 4) {
    temp = check_door(try_doors_x[cc], try_doors_y[cc]);  /* return 1 if door can go there */
    if (temp eq 1) {  /* door can go there */
      try2_x[try2_index] = try_doors_x[cc];
      try2_y[try2_index] = try_doors_y[cc];
      try2_index++;
    }
    cc++;
  }
  /* if the room has no possible doors, than the room intersects both of the
     other rooms - no problem.
     if the room has one or two possible doors, then make all doors.
     if the room has three or four doors, randomly knock out some until there
     are only two doors. */
  while (try2_index > 2) {  /* remove some extra doors */
    remove_this = rnd(try2_index);
    if (remove_this < (try2_index - 1)) {
      try3 = remove_this;
      while (try3 < 3) {
        try2_x[try3] = try2_x[(try3 + 1)];
        try2_y[try3] = try2_y[(try3 + 1)];
        try3++;
      }
    }
    try2_index--;
  }
  /* append door locations to existing global room door array */
  cc = 0;
  while (cc < try2_index) {
    room_doors_x[room_doors_index] = try2_x[cc];
    room_doors_y[room_doors_index] = try2_y[cc];
    cc++;
    room_doors_index++;
  }
}


int write_to_line(l, f)   /* adds a CR for us as well */
char *l;
FILE *f;
{
  int x;
  char *ptr;

  x = strlen(l);
  ptr = l;
  while (x--)
    if (fputc(*(ptr++), f) eq EOF)
      return(DOOPS);
  ptr = (char *) NULL;
  if (fputc('\n', f) eq EOF)
    return(DOOPS);
  return(OKDOKEY);
}


void build_cell(m_x, m_y, dungeon_size)
int m_x;
int m_y;
int dungeon_size;
{
  char temp_str1[4], temp_str2[4];
  char temp_cell_name[MAX_STR];
  int temp_doors_x[4], temp_doors_y[4];
  int status, cc;

  room_doors_index = 0;
  m_str(m_x, &temp_str1[0]);
  m_str(m_y, &temp_str2[0]);
  if (num_levels eq 1) {  /* only one vertical level - shorter different cell name */
    sprintf(  /* build text file name */
      (char *) &txtr_output_file[0],
      "%s%s.txt",
      temp_str1, temp_str2
    );
  } else {  /* multiple vertical levels, longer different cell and file names */
    sprintf(  /* build text file name */
      (char *) &txtr_output_file[0],
      "%s%s%03d.txt",
      temp_str1, temp_str2, (cur_level - 1)
    );
  }
  cell_name(m_x, m_y, &temp_cell_name[0]);  /* build interior cell name */

  /* open output text file for this interior cell */
  if (_txtr_file eq NULL) {
    if ((_txtr_file = fopen((char *) &txtr_output_file[0], "w")) eq NULL) {
      fprintf(stderr, "fopen: error creating output text file %s\n", txtr_output_file);
      exit(DOOPS);
    }
  } else {
    if ((_txtr_file = freopen((char *) &txtr_output_file[0], "w", _txtr_file)) eq NULL) {
      fprintf(stderr, "freopen: error creating output text file %s\n", txtr_output_file);
      exit(DOOPS);
    }
  }
  fprintf(_txtr_file, "reference \"northmarker\" 1 <0 0 0> <0 0 0>\n\n");
  start_point = rnd(4);

  /* extend corridors and put load doors at N, S, E, W exits, put load
     doors w/ doorways or endcaps (nubs) ((dead ends)) instead of
     doors! */
  if (ROAi(m_y) > (1 - ROAi(dungeon_size))) {
    place_straight(0, (maze_size / 2), -1, 0); /* north corridor */
    place_doorjam(90, (maze_size / 2), -1, 0); /* north door */
    place_load_door(90, (maze_size / 2), -1, 0, m_x, m_y, dungeon_size); /* north door */
  } else {
    place_cap(0, (maze_size / 2), -1, 0);
  }
  if (ROAi(m_x) < (ROAi(dungeon_size) - 1)) {
    place_straight(90, maze_size, (maze_size / 2), 0); /* east corridor */
    place_doorjam(180, maze_size, (maze_size / 2), 0); /* east door */
    place_load_door(180, maze_size, (maze_size / 2), 0, m_x, m_y, dungeon_size); /* east door */
  } else {
    place_cap(90, maze_size, (maze_size / 2), 0);
  }
  if (ROAi(m_y) < (ROAi(dungeon_size) - 1)) {
    place_straight(0, (maze_size / 2), maze_size, 0); /* south corridor */
    place_doorjam(270, (maze_size / 2), maze_size, 0); /* south door */
    place_load_door(270, (maze_size / 2), maze_size, 0, m_x, m_y, dungeon_size); /* south door */
  } else {
    place_cap(180, (maze_size / 2), maze_size, 0);
  }
  if (ROAi(m_x) > (1 - ROAi(dungeon_size))) {
    place_straight(90, -1, (maze_size / 2), 0); /* west corridor */
    place_doorjam(0, -1, (maze_size / 2), 0); /* west door */
    place_load_door(0, -1, (maze_size / 2), 0, m_x, m_y, dungeon_size); /* west door */
  } else {
    place_cap(270, -1, (maze_size / 2), 0);
  }


  /* fill maze and walls and rooms with nothing (code 0) */
  xx = 0;
  while (xx < maze_size) {
    yy = 0;
    while (yy < maze_size) {
      maze[xx][yy] = 0;
      rooms[xx][yy] = 0;
      yy++;
    }
    xx++;
  }


  /* keeping track of trap door from level above this one, pick location of
     next trap door down.  Randomly decide if we are even going to put a
     door cell (0, 0) on every level always gets trapdoors */
  if (num_levels > 1) {
    /* save previous levels's door location */
    last_dx = new_dx;
    last_dy = new_dy;
    /* pick new location for door going to next level */
    while ((new_dx eq last_dx) && (new_dy eq last_dy)) {
      new_dx = 1 + rnd((maze_size - 2));
      new_dy = 1 + rnd((maze_size - 2));
    }
    if ((m_x neq 0) || (m_y neq 0)) {
      if (rnd(3)) {
        new_dy = new_dx = -1;  /* no connection to level below */
      }
    }
  }


  /* put first room in */
  x_size = 3;
  y_size = 3;
  x_pos = 2 + rnd((maze_size - x_size - 3));
  y_pos = 2 + rnd((maze_size - y_size - 3));
  room_corner_coords_x[0] = x_pos;
  room_corner_coords_y[0] = y_pos;
  room_corner_coords_x[1] = x_pos + x_size - 1;
  room_corner_coords_y[1] = y_pos + y_size - 1;
  xx = x_pos;
  while (xx < (x_pos + x_size)) {
    yy = y_pos;
    while (yy < (y_pos + y_size)) {
      rooms[xx][yy] = 1;
      yy++;
    }
    xx++;
  }
  room_area = x_size * y_size;


  /* put second room in - it may intersect with the first room */
  x_pos = 2 + rnd((maze_size - x_size - 3));
  y_pos = 2 + rnd((maze_size - y_size - 3));
  room_corner_coords_x[2] = x_pos;
  room_corner_coords_y[2] = y_pos;
  room_corner_coords_x[3] = x_pos + x_size - 1;
  room_corner_coords_y[3] = y_pos + y_size - 1;
  xx = x_pos;
  while (xx < (x_pos + x_size)) {
    yy = y_pos;
    while (yy < (y_pos + y_size)) {
      if (rooms[xx][yy] eq 0) {
        rooms[xx][yy] = 1;
        room_area++;
      }
      yy++;
    }
    xx++;
  }


  /* put third room in - it may intersect with the first two rooms */
  x_pos = 2 + rnd((maze_size - x_size - 3));
  y_pos = 2 + rnd((maze_size - y_size - 3));
  room_corner_coords_x[4] = x_pos;
  room_corner_coords_y[4] = y_pos;
  room_corner_coords_x[5] = x_pos + x_size - 1;
  room_corner_coords_y[5] = y_pos + y_size - 1;
  xx = x_pos;
  while (xx < (x_pos + x_size)) {
    yy = y_pos;
    while (yy < (y_pos + y_size)) {
      if (rooms[xx][yy] eq 0) {
        rooms[xx][yy] = 1;
        room_area++;
      }
      yy++;
    }
    xx++;
  }


  /* pick one starter and dig it out */
  switch (start_point) {
    case 0 :  /* south */
      maze[(maze_size / 2)][(maze_size - 1)] = 4;  /* drill south entrance */
      break;
    case 1 :  /* east */
      maze[(maze_size - 1)][(maze_size / 2)] = 1;  /* drill east entrance */
      break;
    case 2 :  /* west */
      maze[0][(maze_size / 2)] = 2;  /* drill west entrance */
      break;
    default : /* north */
      maze[(maze_size / 2)][0] = 8;  /* drill north entrance - original case */
  }


  /* random part of maze generator */
  cc = ((maze_size - 2) * (maze_size - 2)) - room_area - 5;
  while (cc > 0) {
    try_x = 1 + rnd((maze_size - 2)); /* pick spot */
    try_y = 1 + rnd((maze_size - 2));
    if (maze[try_x][try_y] eq 0) {  /* if spot unfilled */
      if (rooms[try_x][try_y] eq 0) {  /* if no room there */
        filled_neighbor = 0;  /* set to false */
        if (maze[try_x + 1][try_y] > 0) { /* one to the east */
          filled_neighbor = 1;  /* set to true */
          maze[try_x][try_y] = 1;  /* drill east wall */
          maze[try_x + 1][try_y] = maze[try_x + 1][try_y] + 2;  /* drill west wall of east neighbor */
        }
        if (filled_neighbor eq 0) {
          if (maze[try_x - 1][try_y] > 0) { /* one to the west */
            filled_neighbor = 1;  /* set to true */
            maze[try_x][try_y] = 2;  // drill west wall
            maze[try_x - 1][try_y] = maze[try_x - 1][try_y] + 1;  /* drill east wall of west neighbor */
          }
        }
        if (filled_neighbor eq 0) {
          if (maze[try_x][try_y + 1] > 0) {  /* one to the south */
            filled_neighbor = 1;  /* set to true */
            maze[try_x][try_y] = 4;  /* drill south wall */
            maze[try_x][try_y + 1] = maze[try_x][try_y + 1] + 8;  /* drill north wall of south neighbor */
          }
        }
        if (filled_neighbor eq 0) {
          if (maze[try_x][try_y - 1] > 0) { /* one to the north */
            filled_neighbor = 1;  /* true */
            maze[try_x][try_y] = 8;  /* drill north wall */
            maze[try_x][try_y - 1] = maze[try_x][try_y - 1] + 4;  /* drill south wall of north neighbor */
          }
        }
        if (filled_neighbor eq 1) { /* if true */
          cc--;
        }
      }
    } /* end if */
  }


  /* non random part
     drill any points that were not found during the random part */
  xx = 1;
  while (xx < (maze_size - 1)) {
    yy = 1;
    while (yy < (maze_size - 1)) {
      if (rooms[xx][yy] eq 0) {  /* if not a room */
        if (maze[xx][yy] eq 0) {  /* if undrilled / undug */
          filled_neighbor = 0;  /* set to false */
          if (maze[xx + 1][yy] > 0) {  /* one to the east */
            filled_neighbor = 1;  /* true */
            maze[xx][yy] = 1;  /* drill east wall */
            maze[xx + 1][yy] = maze[xx + 1][yy] + 2;  /* drill west wall of east neighbor */
          }
          if (filled_neighbor eq 0) { /* if flase */
            if (maze[xx - 1][yy] > 0) { /* one to the west */
              filled_neighbor = 1;  /* true */
              maze[xx][yy] = 2;  /* drill west wall */
              maze[xx - 1][yy] = maze[xx - 1][yy] + 1;  /* drill east wall of west neighbor */
            }
          }
          if (filled_neighbor eq 0) {  /* if false */
            if (maze[xx][yy + 1] > 0) {  /* one to the south */
              filled_neighbor = 1;  /* true */
              maze[xx][yy] = 4;  /* drill south wall */
              maze[xx][yy + 1] = maze[xx][yy + 1] + 8;  /* drill north wall of south neighbor */
            }
          }
          if (filled_neighbor eq 0) {  /* if false */
            if (maze[xx][yy - 1] > 0) {  /* one to the north */
              filled_neighbor = 1;
              maze[xx][yy] = 8;  /* drill north wall */
              maze[xx][yy - 1] = maze[xx][yy - 1] + 4;  /* drill south wall of north neighbor */
            }
          }
        }
      }
      yy = yy + 1;
    }
    xx = xx + 1;
  }


  /* drill remaining hardcoded exit points (s, w, e) */
  /* west exit */
  if (start_point != 2) {
    maze[0][(maze_size / 2)] = 3;
    maze[1][(maze_size / 2)] = maze[1][(maze_size / 2)] + 2;
  }
  /* east exit */
  if (start_point != 1) {
    maze[(maze_size - 1)][(maze_size / 2)] = 3;
    maze[(maze_size - 2)][(maze_size / 2)] = maze[(maze_size - 2)][(maze_size / 2)] + 1;
  }
  /* south exit */
  if (start_point != 0) {
    maze[(maze_size / 2)][(maze_size - 1)] = 12;
    maze[(maze_size / 2)][(maze_size - 2)] = maze[(maze_size / 2)][(maze_size - 2)] + 4;
  }
  /* north exit */
  if (start_point != 3) {
    maze[(maze_size / 2)][0] = 12;
    maze[(maze_size / 2)][1] = maze[(maze_size / 2)][1] + 8;
  }


  /* determine where room doors will go */
  /* first room (of three) */
  temp_doors_x[0] = room_corner_coords_x[0] + 1; /* north */
  temp_doors_y[0] = room_corner_coords_y[0];  /* north */
  temp_doors_x[1] = room_corner_coords_x[0];  /* west */
  temp_doors_y[1] = room_corner_coords_y[0] + 1;  /* west */
  temp_doors_x[2] = room_corner_coords_x[1];  /* east */
  temp_doors_y[2] = room_corner_coords_y[1] - 1;  /* east */
  temp_doors_x[3] = room_corner_coords_x[1] - 1;  /* south */
  temp_doors_y[3] = room_corner_coords_y[1];  /* south */
  find_doors(
    &temp_doors_x[0], &temp_doors_y[0]
  );
  /* second room (of three) */
  temp_doors_x[0] = room_corner_coords_x[2] + 1; /* north */
  temp_doors_y[0] = room_corner_coords_y[2];  /* north */
  temp_doors_x[1] = room_corner_coords_x[2];  /* west */
  temp_doors_y[1] = room_corner_coords_y[2] + 1;  /* west */
  temp_doors_x[2] = room_corner_coords_x[3];  /* east */
  temp_doors_y[2] = room_corner_coords_y[3] - 1;  /* east */
  temp_doors_x[3] = room_corner_coords_x[3] - 1;  /* south */
  temp_doors_y[3] = room_corner_coords_y[3];  /* south */
  find_doors(
    &temp_doors_x[0], &temp_doors_y[0]
  );
  /* third room (of three) */
  temp_doors_x[0] = room_corner_coords_x[4] + 1; /* north */
  temp_doors_y[0] = room_corner_coords_y[4];  /* north */
  temp_doors_x[1] = room_corner_coords_x[4];  /* west */
  temp_doors_y[1] = room_corner_coords_y[4] + 1;  /* west */
  temp_doors_x[2] = room_corner_coords_x[5];  /* east */
  temp_doors_y[2] = room_corner_coords_y[5] - 1;  /* east */
  temp_doors_x[3] = room_corner_coords_x[5] - 1;  /* south */
  temp_doors_y[3] = room_corner_coords_y[5];  /* south */
  find_doors(
    &temp_doors_x[0], &temp_doors_y[0]
  );
  /* if two rooms have identical or coords, there could be
     a duplicate door */
  cc = 0;
  while (cc < room_doors_index) {
    skip_this_door = 0; /* false */
    if (rooms[room_doors_x[cc] - 1][room_doors_y[cc]] eq 2) {
      skip_this_door = 1; /* true */
    }
    if (rooms[room_doors_x[cc] + 1][room_doors_y[cc]] eq 2) {
      skip_this_door = 1; /* true */
    }
    if (rooms[room_doors_x[cc]][room_doors_y[cc] - 1] eq 2) {
      skip_this_door = 1;  /* true */
    }
    if (rooms[room_doors_x[cc]][room_doors_y[cc] + 1] eq 2) {
      skip_this_door = 1;  /* true */
    }
    if (skip_this_door eq 0) {  /* if false */
      if (rooms[room_doors_x[cc]][room_doors_y[cc]] eq 1) {
        /* 1 - room no door yet, 0 = no room, 2 = room with door */
        rooms[room_doors_x[cc]][room_doors_y[cc]] = 2;
        if (rooms[room_doors_x[cc]][room_doors_y[cc] - 1] eq 0) { /* door in room north wall */
          maze[room_doors_x[cc]][room_doors_y[cc] - 1] =
          maze[room_doors_x[cc]][room_doors_y[cc] - 1] + 4; /* drill corridor south side */
        } else {
          if (rooms[room_doors_x[cc]][room_doors_y[cc] + 1] eq 0) { /* door in room south wall */
            maze[room_doors_x[cc]][room_doors_y[cc] + 1] =
            maze[room_doors_x[cc]][room_doors_y[cc] + 1] + 8; /* drill corridor north side */
          } else {
            if (rooms[room_doors_x[cc] - 1][room_doors_y[cc]] eq 0) {  /* door in room west wall */
              maze[room_doors_x[cc] - 1][room_doors_y[cc]] =
              maze[room_doors_x[cc] - 1][room_doors_y[cc]] + 1; /* drill corridor east side */
            } else {  /* door in room east wall */
              maze[room_doors_x[cc] + 1][room_doors_y[cc]] =
              maze[room_doors_x[cc] + 1][room_doors_y[cc]] + 2; /* drill corridor west side */
            }
          }
        }
      }
    }
    cc++;
  }


  /* nub-removal */
  how_twisty = rnd(4);  /* 0 is max twistyness (zero nub removal) */
  if (how_twisty > 2) {
    how_twisty = 2;
  }
  if (no_twisty) {  /* command line switch for no twistyness */
    how_twisty = 2;
  }
  cc = 0;
  while (cc < how_twisty) {
    yy = 1;
    while (yy < (maze_size - 1)) {
      xx = 1;
      while (xx < (maze_size - 1)) {
        leave_it = 0;  /* false */
        /* if there is a non-diagonally adjacent room entrance, LEAVE IT ALONE */
        if (rooms[(xx - 1)][yy] eq 2) {  /* 2 is code for room entrance */
          leave_it = 1;  /* true */
        }
        if (rooms[(xx + 1)][yy] eq 2) {
          leave_it = 1;  /* true */
        }
        if (rooms[xx][(yy - 1)] eq 2) {
          leave_it = 1;  /* true */
        }
        if (rooms[xx][(yy + 1)] eq 2) {
          leave_it = 1;  /* true */
        }
        /* if there is a trap door from the room above, don't remove this nub */
        if (cur_level > 1) {  /* if there is a level above this level */
          if ((xx eq last_dx) && (yy eq last_dy)) {
            leave_it = 1;  /* true */
          }
        }
        /* if there is a trap door to the room below, don't remove this nub */
        if (cur_level < num_levels) { /* if there is a level below this level */
          if ((xx eq new_dx) && (yy eq new_dy)) {
            leave_it = 1;  /* true */
          }
        }
        /* if leave_it is still false, remove the nub */
        if (leave_it eq 0) {  /* if false */
          if (maze[xx][yy] eq 8) { /* north open nub */
            maze[xx][yy] = 0; /* un-drill */
            maze[xx][yy - 1] = maze[xx][yy - 1] - 4;  /* restore south wall */
          } else {
            if (maze[xx][yy] eq 4) {  /* south open nub */
              maze[xx][yy] = 0;  /* un-drill */
              maze[xx][yy + 1] = maze[xx][yy + 1] - 8;  /* restore north wall */
            } else {
              if (maze[xx][yy] eq 2) {  /* west open nub */
                maze[xx][yy] = 0;  /* un-drill */
                maze[xx - 1][yy] = maze[xx - 1][yy] - 1;  /* restore east wall */
              } else {
                if (maze[xx][yy] eq 1) { /* east open nub */
                  maze[xx][yy] = 0; /* un-drill */
                  maze[xx + 1][yy] = maze[xx + 1][yy] - 2;  /* restore west wall */
                }
              }
            }
          }
        } /* end if leave_it */
        xx++;
      }
      yy++;
    }
    cc++;
  }


  /* display (output) the maze */
  /* m_center = (maze_size / 2); */
  xx = 0;
  while (xx < maze_size) {
    yy = 0;
    while (yy < maze_size) {
      if (maze[xx][yy] > 0) {  /* corridor */
        place_corridor(xx, yy, 0);
      } else {
        if (rooms[xx][yy] > 0) {
          place_room(xx, yy, 0);
        }
      }
      yy++;
    }
    xx++;
  }

  /* write trap door up (if there is one) */
  if (cur_level > 1) {
    if (last_dx neq -1) {
      /* rebuild cell name to point to cell above - modify cur_level */
      cur_level--;
      cell_name(m_x, m_y, &temp_cell_name[0]);  /* build interior cell name */
      cur_level++;
      /* write the door & the pink arrow */
      fprintf(
        _txtr_file,
        "tel_door_ref \"in_com_traptop_01\" 1 <0 180 0> <%d %d 110> <0 0 %d> <%d %d 0> \"%s\"\n",
        esp_grid(last_dx), 0 - esp_grid(last_dy),
        /* pink arrow stuff */
        90 * rnd(4),  /* pink arrow rot */
        esp_grid(last_dx), 0 - esp_grid(last_dy),
        temp_cell_name
      );
    }
  }

  /* write trap door down (if there is one) */
  if (cur_level < num_levels) {
    if (new_dx neq -1) {
      /* rebuild cell name to point to cell above - modify cur_level */
      cur_level++;
      cell_name(m_x, m_y, &temp_cell_name[0]);  /* build interior cell name */
      cur_level--;
      /* write the door and the pink arrow */
      fprintf(
        _txtr_file,
        "tel_door_ref \"in_com_traptop_01\" 1 <0 0 0> <%d %d -131> <0 0 %d> <%d %d 0> \"%s\"\n",
        esp_grid(new_dx), 0 - esp_grid(new_dy),
        /* pink arrow stuff */
        90 * rnd(4),  /* pink arrow rot */
        esp_grid(new_dx), 0 - esp_grid(new_dy),
        temp_cell_name
      );
    }
  }

  fclose(_txtr_file);
}


int write_bat_file(bat_file_name, dun_size, base_file_name)
char *bat_file_name;
int dun_size;
char *base_file_name;
{
  int m_x, m_y, status;
  char temp_str1[4], temp_str2[4];
  char txt_file_name[MAX_STR];
  char temp_cell_name[MAX_STR];
  char bat_temp[MAX_STR];
  FILE *_bat_file, *_list_file;

  _bat_file = fopen(bat_file_name, "w");
  if (_bat_file eq NULL) {
    fprintf(stderr, "error creating bat file \"%s\"\n", bat_file_name);
    exit(DOOPS);
  }
  _list_file = fopen(out_list_file_name, "w");
  if (_list_file eq NULL) {
    fprintf(stderr, "error creating list file \"%s\"\n", out_list_file_name);
    exit(DOOPS);
  }

  /* write include esp to list file */
  if (include_esp_file_name[0] neq '\0') {  /* if there is an include esp */
    if (0 > fprintf(_list_file, "%s\n", &include_esp_file_name[0])) {
      fprintf(stderr, "error writing first line to list file\n");
      fprintf(stderr, "errno is: %d\n", errno);
      exit(DOOPS);
    }
  }


  m_y = (1 - ROAi(dun_size));
  while (ROAi(m_y) < ROAi(dun_size)) {

    m_x = (1 - ROAi(dun_size));
    while (ROAi(m_x) < ROAi(dun_size)) {

      cur_level = 1;
      while (ROAi(cur_level) <= ROAi(num_levels)) {
        if (ROAi(m_x) < ROAi(dun_size)) {
          if (ROAi(m_y) < ROAi(dun_size)) {
            m_str(ROAi(m_x), &temp_str1[0]);
            m_str(ROAi(m_y), &temp_str2[0]);
            if (num_levels eq 1) { /* one level */
              sprintf(  /* build txt file name (less the .txt extension) */
                (char *) &txt_file_name[0],
                "%s%s.",
                temp_str1, temp_str2
              );
            } else {  /* multiple levels */
              sprintf(  /* build txt file name (less the .txt extension) */
                (char *) &txt_file_name[0],
                "%s%s%03d.",
                temp_str1, temp_str2, (cur_level - 1)
              );
            }
            cell_name(ROAi(m_x), ROAi(m_y), &temp_cell_name[0]);  /* build interior cell name */
            sprintf(
              &bat_temp[0],
              "txt2esp4 %stxt %sesp \"%s\"",
              txt_file_name,  /* string containing name of txt file */
              txt_file_name,  /* also used for esp file name */
              temp_cell_name  /* name of interior cell (as it will appear in the game) */
            );

            /* write to bat file */
            if (0 > fprintf(_bat_file, "%s\n", &bat_temp[0])) {
              fprintf(stderr, "error writing line to bat file\n");
              fprintf(stderr, "errno is: %d\n", errno);
              exit(DOOPS);
            }

            /* write to list file */
            if (0 > fprintf(_list_file, "%sesp\n", &txt_file_name[0])) {
              fprintf(stderr, "error writing line to list file\n");
              fprintf(stderr, "errno is: %d\n", errno);
              exit(DOOPS);
            }

          }
        }

        cur_level++;
      } /* end while */

      m_x++;
    } /* end of while m_x */

    m_y++;
  } /* end while m_y */

  /* write espjoin line to bat file */
  if (0 > fprintf(_bat_file, "espjoin %s %s.esp\n", out_list_file_name, base_file_name)) {
    fprintf(stderr, "error writing last espjoin to bat file\n");
    fprintf(stderr, "errno is: %d\n", errno);
    exit(DOOPS);
  }

  /* delete temp files (except for .bat and list.txt file) */
  if (del_files) {
    m_y = (1 - ROAi(dun_size));
    while (ROAi(m_y) < ROAi(dun_size)) {
      m_x = (1 - ROAi(dun_size));
      while (ROAi(m_x) < ROAi(dun_size)) {
        cur_level = 1;
        while (ROAi(cur_level) <= ROAi(num_levels)) {
          if (ROAi(m_x) < ROAi(dun_size)) {
            if (ROAi(m_y) < ROAi(dun_size)) {
              m_str(ROAi(m_x), &temp_str1[0]);
              m_str(ROAi(m_y), &temp_str2[0]);
              if (num_levels eq 1) { /* one level */
                sprintf(  /* build esp file name (less the .esp extension) */
                  (char *) &txt_file_name[0],
                  "%s%s.",
                  temp_str1, temp_str2
                );
              } else {  /* multiple levels */
                sprintf(  /* build esp file name (less the .esp extension) */
                  (char *) &txt_file_name[0],
                  "%s%s%03d.",
                  temp_str1, temp_str2, (cur_level - 1)
                );
              }
              /* write del .txt command to bat file */
              if (0 > fprintf(_bat_file, "del %stxt\n", txt_file_name)) {
                fprintf(stderr, "error writing line to bat file\n");
                fprintf(stderr, "errno is: %d\n", errno);
                exit(DOOPS);
              }
              /* write del .esp command to bat file */
              if (0 > fprintf(_bat_file, "del %sesp\n", txt_file_name)) {
                fprintf(stderr, "error writing line to bat file\n");
                fprintf(stderr, "errno is: %d\n", errno);
                exit(DOOPS);
              }
            }
          }
          cur_level++;
        }  /* end while */
        m_x++;
      }  /* end while */
      m_y++;
    }  /* end while */
  }  /* end if del_file */

  fclose(_list_file);
  fclose(_bat_file);
  return(OKDOKEY);
}


int parse_options(argc, argv)  /* parse command line switches, if any */
int argc;
char *argv[];
{
  int xx;

  xx = 4;
  no_treasure = no_monsters = no_clutter = no_twisty = 0;
  num_levels = 1;
  lev_offset = 0;
  while (xx < argc) {
    if (0 eq stricmp(argv[xx], "-t")) {
      no_treasure = 1;
    } else {
      if (0 eq stricmp(argv[xx], "-m")) {
        no_monsters = 1;
      } else {
        if (0 eq stricmp(argv[xx], "-c")) {
          no_clutter = 1;
        } else {
          if (0 eq stricmp(argv[xx], "-nt")) { /* no twistiness */
            no_twisty = 1;
          } else {
            if (0 eq stricmp(argv[xx], "-l")) { /* set number of (vertical) levels */
              xx++;
              if (xx >= argc) {
                fprintf(stderr, "\"-l\" option must be followed by a space");
                fprintf(stderr, " and then by a positive integer\n");
                return(DOOPS);
              }
              num_levels = atoi(argv[xx]);
              if (num_levels < 1) {
                fprintf(stderr, "\"-l\" option must be followed by a space");
                fprintf(stderr, " and then by a positive integer\n");
                return(DOOPS);
              }
              if (num_levels > 1000) {
                fprintf(stderr, "too many levels.  1000 is the upper limit\n");
                return(DOOPS);
              }
              if ((num_levels + lev_offset) > 99999999) {
                fprintf(stderr, "combined number of levels and level offset exceeds 99999999\n");
                return(DOOPS);
              }
            } else {
              if (0 eq stricmp(argv[xx], "-sl")) {  /* start level */
                xx++;
                if (xx >= argc) {
                  fprintf(stderr, "\"-sl\" option must be followed by a space");
                  fprintf(stderr, " and then by a positive integer\n");
                  return(DOOPS);
                }
                lev_offset = atoi(argv[xx]);
                if (lev_offset < 1) {
                  fprintf(stderr, "\"-sl\" option must be followed by a space");
                  fprintf(stderr, " and then by a positive integer\n");
                  return(DOOPS);
                }
                lev_offset--;
                if ((num_levels + lev_offset) > 99999999) {
                  fprintf(stderr, "combined number of levels and level offset exceeds 99999999\n");
                  return(DOOPS);
                }
              } else {
                fprintf(stderr, "bad command line switch \"%s\"\n", argv[xx]);
                return(DOOPS);
              }
            }
          }
        }
      }
    }
    xx++;
  } 
  return(OKDOKEY);
}


main(argc, argv)
int argc;
char *argv[];
{
  int x, status, cc;
  int m_x, m_y, dungeon_size;
  char temp_cell_name[MAX_STR];
  char temp_str_long[MAX_STR], bat_temp[MAX_STR];

  if (argc eq 0)
    exit(DOOPS);

  if (argc < 4)  usage_exit();
  strcpy(base_file_name, argv[1]);
  dungeon_size = atoi(argv[2]);
  if ((ROAi(dungeon_size )< 0) || (ROAi(dungeon_size) > 99)) {
    fprintf(stderr, "Error.  dungeon size must be 1-99 inclusive\n");
    exit(DOOPS);
  }
  strcpy(base_cell_name, argv[3]);
  for (x = 0; x neq 20; x++) {
    leveled_monster[x] = (char *) NULL;
    crate[x] = (char *) NULL;
    bag[x] = (char *) NULL;
    small_chest[x] = (char *) NULL;
  }
  if ((status = parse_options(argc, argv)) eq DOOPS) {
    exit(DOOPS);
  }
  /* next: handle/kludge the special case of a level offset
           combined with only one level */
  if (lev_offset > 0) {
    if (num_levels eq 1) {
      sprintf(base_cell_name, "%s, Level %d", argv[3], (1 + lev_offset) );
    }
  }
  leveled_monster_index = 0;
  crate_index = 0;
  bag_index = 0;
  small_chest_index = 0;
  include_esp_file_name[0] = '\0';
  out_bat_file_name[0] = '\0';
  out_list_file_name[0] = '\0';
  if ((status = parse_ini()) neq OKDOKEY) {
    fprintf(stderr, "nonfatal warning: error parsing impgen.ini file\n");
  }
  /* some or all values are blank and need to be set to defaults */
  fill_defaults();

  set_seed();  /* simple way to initialize random seed */

  m_y = (1 - ROAi(dungeon_size));
  while (ROAi(m_y) < ROAi(dungeon_size)) {

    m_x = (1 - ROAi(dungeon_size));
    while (ROAi(m_x) < ROAi(dungeon_size)) {

      cur_level = 1;
      while (cur_level <= num_levels) {

        if (ROAi(m_x) < ROAi(dungeon_size))
          if (ROAi(m_y) < ROAi(dungeon_size))
            build_cell(ROAi(m_x), ROAi(m_y), ROAi(dungeon_size));
        /* done creating the cell */

        cur_level++;
      }  /* end of while cur_level */

      m_x++;
    } /* end of while m_x */

    m_y++;
  } /* end while m_y */

  /* almost done - write bat file now */
  status = write_bat_file(out_bat_file_name, ROAi(dungeon_size), base_file_name);
  if (status neq OKDOKEY) {
    fprintf(stderr, "error writing bat file\n");
    exit(DOOPS);
  }

  update_seed();
  exit(OKDOKEY);
}


/* actual end of this file */
